/*
* ===========================================
* Java Pdf Extraction Decoding Access Library
* ===========================================
*
* Project Info:  http://www.jpedal.org
* Project Lead:  Mark Stephens (mark@idrsolutions.com)
*
* (C) Copyright 2003, IDRsolutions and Contributors.
* 
* 	This file is part of JPedal
*
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


*
* ---------------
* PdfData.java
* ---------------
* (C) Copyright 2002, by IDRsolutions and Contributors.
*
* Original Author:  Mark Stephens (mark@idrsolutions.com)
* Contributor(s):
*/
package org.jpedal.objects;

import java.util.*;

import org.jpedal.PdfDecoder;
import org.jpedal.color.GenericColorSpace;
import org.jpedal.utils.Fonts;
import org.jpedal.utils.repositories.Vector_Int;


/**
 * <p>
 * holds text data for extraction & manipulation
 * </p>
 * <p>
 * Pdf routines create 'raw' text data
 * </p>
 * <p>
 * grouping routines will attempt to intelligently stitch together and leave as
 * 'processed data' in this class
 * </p>
 * <p>
 * <b>NOTE ONLY methods (NOT public variables) are part of API </b>
 * </p>
 * We then transfer the data into our generic Storypad object store (which also
 * serves as an examaple of how the data can be used).
 *  
 */
public class PdfData extends StoryData
{

    /**identify type as text*/
    public static final int TEXT = 0;

    /**identify type as image*/
    public static final int IMAGE = 1;

    /**
     * list of elements - each element is a text fragment from the pdf
     */
    private java.util.List  text_objects;
    
    /**holds fields extractred by regexp*/
    Map extractedTextFields=new HashMap(),extractedXMLFields=new HashMap();

    private int pointer=0;

    private StringBuffer markedContent=null;

    /**holds items we used to create the story*/
    public String[] fragments=new String[max];

    public int[] parent=new int[max];

    /**flag to show x co-ord has been embedded in content*/
    private boolean widthIsEmbedded=false;

    /**test orientation*/
    public static final int HORIZONTAL_LEFT_TO_RIGHT = 0;

    public static final int HORIZONTAL_RIGHT_TO_LEFT = 1;

    public static final int VERTICAL_TOP_TO_BOTTOM = 2;

    public static final int VERTICAL_BOTTOM_TO_TOP = 3;

    /**local store for max and widthheight of page*/
    public float maxY=0,maxX=0;

    public int[] fontSize;

    /**vlaues same for all records in file*/
    private Map globalValues=new HashMap();

    /**flagged if marked content read*/
	private boolean isMarked=false;

    /** create empty object to hold content*/
    public PdfData()
    {
        text_objects = new Vector();

        //set all parent values
        for(int i=0;i<max;i++)
            parent[i]=-1;
    }

    /**
     * get data as a Map
     */
    final public Map getTextElementAt( int i )
    {
        return (Map)text_objects.get( i );
    }


    /**
     * get number of raw objects on page
     */
    final public int getRawTextElementCount()
    {

        return pointer;
    }

    /**
     * reset values
     */
    final public void resetTextList( java.util.List new_text_objects )
    {
        text_objects = new_text_objects;


    }

    //////////////////////////////////////////////////////////////////
    /**
     * clear store of objects once written out
     * to reclaim memory. If flag set, sets data to
     * state after page decoded before grouping for reparse
     */
    final public void flushTextList( boolean reinit )
    {
        text_objects=new Vector();

        if( reinit == false ){

            pointer=0;

            contents=new String[max];
            f_writingMode=new int[max];
            f_font_used=new String[max];
            font_data=new String[max];
            text_length=new int[max];
            move_command=new int[max];
            f_character_spacing=new float[max];
            token_counter=new int[max];
            f_end_font_size=new int[max];
            space_width=new float[max];
            f_x1=new float[max];
            colorTag=new String[max];
            f_x2=new float[max];
            f_y1=new float[max];
            f_y2=new float[max];

        }
    }

    /////////////////////////////////////////////////////////////////
    /**
     * get number of objects at end
     */
    final public int getTextElementCount()
    {
        return text_objects.size();
    }

    /**
     * store line of raw text for later processing
     */
    final public void addImageElement( float x1, float y1, float width, float height,
                                       String imageName){

        f_x1[pointer]=x1;
        objectType[pointer]=IMAGE;
        f_x2[pointer]=x1+width;
        f_y1[pointer]=y1+height;
        f_y2[pointer]=y1;
        fragments[pointer]=""+pointer;
        contents[pointer]=imageName;
        unformattedContent[pointer]=imageName;

        pointer++;

        //resize pointers
        if(pointer==max){
            resizeArrays();
        }
        
    }

    /**
     * store line of raw text for later processing
     */
    final public void addRawTextElement( float character_spacing,int writingMode,
                                         String font_as_string,  float current_space, TextState current_text_state,
                                         float x1, float y1, float x2, float y2, int move_type,
                                         StringBuffer processed_line, int token_number, int current_text_length, String currentColorTag )
    {

        //if( ( x1 >= 0 ) && ( y1 >= 0 ) && ( x2 > 0 ) && ( y2 > 0 ) && (processed_line.length()>0)  )
        if(processed_line.length()>0){

            //add fonts to processed data - duplicates stripped out at end
            //if(PdfDecoder.isXMLExtraction())
            //processed_line = translateCharacters( processed_line );

            //add tokens
            if(PdfDecoder.isXMLExtraction()){
                processed_line.insert( 0, font_as_string );
                processed_line.append( Fonts.fe );
            }

            //add color token
            if(isColorExtracted()){
                processed_line.insert( 0, currentColorTag );
                processed_line.append( GenericColorSpace.ce );
            }

            f_writingMode[pointer]=writingMode;
            font_data[pointer]=font_as_string;
            text_length[pointer]=current_text_length;
            move_command[pointer]=move_type;
            f_character_spacing[pointer]=character_spacing;
            f_x1[pointer]=x1;
            colorTag[pointer]=currentColorTag;
            f_x2[pointer]=x2;
            f_y1[pointer]=y1;
            f_y2[pointer]=y2;
            objectType[pointer]=TEXT;
            //raw_data[pointer]=commands_processed.toString();
            contents[pointer]=processed_line.toString();
            token_counter[pointer]=token_number;

            int font_size = current_text_state.getCurrentFontSize();
            f_end_font_size[pointer]=font_size;

            f_font_used[pointer]=current_text_state.getFontName();

            space_width[pointer]=current_space*1000;

            pointer++;

            //resize pointers
            if(pointer==max)
                resizeArrays();
        }
    }

    /**
     * store line of raw text for later processing
     */
    final public void addRawBrokenFragment(int j, StringBuffer text, float x1,float x2,
                                           float y1,float y2,String plotString ,String currentColorTag){

        //make sure valid positive co-ords
        if( ( x1 > 0 ) & ( y1 > 0 ) & ( x2 > 0 ) & ( y2 > 0 ) )
        {

            //add fonts to processed data - duplicates stripped out at end
            //processed_line = translateCharacters( processed_line );

            //add token to close
            if(PdfDecoder.isXMLExtraction()){
                if((!text.toString().endsWith(Fonts.fe))&&
                        (!text.toString().endsWith(GenericColorSpace.ce)))
                    text.append(Fonts.fe );
            }

            f_writingMode[pointer]=f_writingMode[j];
            font_data[pointer]=font_data[j];
            text_length[pointer]=text_length[j];
            move_command[pointer]=move_command[j];
            f_character_spacing[pointer]=f_character_spacing[j];
            f_x1[pointer]=x1;
            f_x2[pointer]=x2;
            f_y1[pointer]=y1;
            f_y2[pointer]=y2;
            colorTag[pointer]=currentColorTag;

            //System.out.println("broken before="+contents[pointer]);
            contents[pointer]=text.toString();
            token_counter[pointer]=token_counter[j];

            f_end_font_size[pointer]=f_end_font_size[j];
            f_font_used[pointer]=f_font_used[j];
            space_width[pointer]=space_width[j];

            pointer++;

            //resize pointers
            if(pointer==max)
                resizeArrays();
        }
    }
    /**
     * 
     */
    private void resizeArrays() {
        float[] temp_f;
        int[] temp_i;
        boolean[] temp_b;
        String[] temp_s;

        if(max<5000)
            max=max*5;
        else if(max<10000)
            max=max*2;
        else
            max=max+1000;

        temp_s=contents;
        contents = new String[max];
        System.arraycopy( temp_s, 0, contents, 0, pointer );

        temp_s=unformattedContent;
        unformattedContent = new String[max];
        System.arraycopy( temp_s, 0, unformattedContent, 0, pointer );

        temp_i=f_writingMode;
        f_writingMode=new int[max];
        f_writingMode = new int[max];
        System.arraycopy( temp_i, 0, f_writingMode, 0, pointer );

        temp_s=f_font_used;
        f_font_used = new String[max];
        System.arraycopy( temp_s, 0, f_font_used, 0, pointer );

        temp_s=colorTag;
        colorTag = new String[max];
        System.arraycopy( temp_s, 0, colorTag, 0, pointer );

        temp_s=font_data;
        font_data = new String[max];
        System.arraycopy( temp_s, 0, font_data, 0, pointer );

        temp_s=fragments;
        fragments = new String[max];
        System.arraycopy( temp_s, 0, fragments, 0, pointer );

        temp_i=text_length;
        text_length = new int[max];
        System.arraycopy( temp_i, 0, text_length, 0, pointer );

        temp_i=move_command;
        move_command = new int[max];
        System.arraycopy( temp_i, 0, move_command, 0, pointer );

        temp_f=f_character_spacing;
        f_character_spacing = new float[max];
        System.arraycopy( temp_f, 0, f_character_spacing, 0, pointer );

        temp_i=parent;
        parent = new int[max];
        //set to default -1
        for(int ii=pointer;ii<max;ii++)
            parent[ii]=-1;
        System.arraycopy( temp_i, 0, parent, 0, pointer );

        temp_s=prefix;
        prefix = new String[max];
        //set to default -1
        for(int ii=pointer;ii<max;ii++)
            prefix[ii]=null;
        System.arraycopy( temp_s, 0, prefix, 0, pointer );

        temp_i=token_counter;
        token_counter = new int[max];
        System.arraycopy( temp_i, 0, token_counter, 0, pointer );

        Object[] temp_o=links;
        links = new Object[max];
        System.arraycopy( temp_o, 0, links, 0, pointer );

        temp_s=category;
        category = new String[max];
        System.arraycopy( temp_s, 0, category, 0, pointer );

        temp_i=objectType;
        objectType = new int[max];
        System.arraycopy( temp_i, 0, objectType, 0, pointer );


        temp_i=f_end_font_size;
        f_end_font_size = new int[max];
        System.arraycopy( temp_i, 0, f_end_font_size, 0, pointer );

        temp_f=space_width;
        space_width = new float[max];
        System.arraycopy( temp_f, 0, space_width, 0, pointer );

        temp_f=f_x1;
        f_x1 = new float[max];
        System.arraycopy( temp_f, 0, f_x1, 0, pointer );

        temp_f=f_x2;
        f_x2 = new float[max];
        System.arraycopy( temp_f, 0, f_x2, 0, pointer );

        temp_f=f_y1;
        f_y1 = new float[max];
        System.arraycopy( temp_f, 0, f_y1, 0, pointer );

        temp_f=f_y2;
        f_y2 = new float[max];
        System.arraycopy( temp_f, 0, f_y2, 0, pointer );
    }

    /**
     * remove a Text Object
     */
    final public void removeTextElementAt( int counter )
    {
        text_objects.remove( new Integer(counter) );

    }

    /**
     * return id of raw fragments stored as space deliminated string
     */
    public String getFragments(int i) {

        return fragments[i];
    }

    /**
     * get parent item for linked items
     */
    public int getParent(int item_selected) {
        return parent[item_selected];
    }

    /**
     * set prefix item for linked items
     */
    public String getPrefix(int item_selected) {
        return prefix[item_selected];
    }

    /**
     * set prefix
     */
    public void setPrefix(int item, String prefixValue){

        //resize if needed
        int size=prefix.length;
        if(item>=size){
            max=prefix.length+100;
            String[] temp_i=prefix;
            prefix = new String[max];
            System.arraycopy( temp_i, 0, prefix, 0, size );

            //fill up gaps with -1
            for(int j=size;j<max;j++)
                prefix[j]=null;
        }

        prefix[item]=prefixValue;
    }

    /**
     * set parent
     */
    public void setParent(int item, int parentID){
        System.out.println("set parent="+item+" "+parentID);
        //resize if needed
        int size=parent.length;
        if(item>=size){
            max=parent.length+100;
            int[] temp_i=parent;
            parent = new int[max];
            System.arraycopy( temp_i, 0, parent, 0, size );

            //fill up gaps with -1
            for(int j=size;j<max;j++)
                parent[j]=-1;
        }

        parent[item]=parentID;
    }

    /**
     * set flag to show width in text
     */
    public void widthIsEmbedded() {

        widthIsEmbedded=true;

    }

    /**
     * show if width in text
     */
    public boolean IsEmbedded() {

        return widthIsEmbedded;

    }
    
    /**fonts size of object - if it changes in object will be FIRST value*/
    public int getFontSize(int currentStory) {
        return fontSize[currentStory];
    }

    /**tag font at start of object*/
    public String getFontTag(int currentStory) {
        return f_font_used[currentStory];
    }

    public String getGlobalField(String globalField) {


        return (String) globalValues.get(globalField);
    }

    public void setGlobalField(String globalNameForCategory,String value) {
        globalValues.put(globalNameForCategory,value);
    }


    public Object[] getGlobalFields() {
        return globalValues.keySet().toArray();
    }

    public int getGlobalCategoryCount() {
        return globalValues.size();
    }

    /**
     * flag set when regexp applied
     */
    public int getRegExpStatus(int i) {

        return regExpStatus[i];
    }

    /**
     * set regexp status
     */
    public void setRegExpStatus(int item, int value){

        //resize if needed
        int size=regExpStatus.length;
        if(item>=size){
            max=regExpStatus.length+100;
            if(item>max)
            	max=item+20;
            int[] temp_i=regExpStatus;
            regExpStatus = new int[max];
            System.arraycopy( temp_i, 0, regExpStatus, 0, size );

            //fill up gaps with -1
            for(int j=size;j<max;j++)
                regExpStatus[j]=-1;
        }

        regExpStatus[item]=value;
    }


	public void addExtractedXMLField(String fieldName, String data,int id) {
		
		Integer key=new Integer(id);
		
		//get existing value and create if needed
		Map currentExtractedFields=(Map) extractedXMLFields.get(key);
		if(currentExtractedFields==null)
			currentExtractedFields=new HashMap();
		
		//add to existing value and store
		String currentValue=(String) currentExtractedFields.get(fieldName);
		if(currentValue!=null)
			data=currentValue+data;
		System.out.println("add "+fieldName);
		currentExtractedFields.put(fieldName,data);
		
		//save at end
		extractedXMLFields.put(key,currentExtractedFields);
		
	}

	public Map getExtractedTextFields(int id) {
		
		return (Map) extractedTextFields.get(new Integer(id));
	}
	
	public Map getExtractedXMLFields(int id) {
		
		return (Map) extractedXMLFields.get(new Integer(id));
	}
	
	public void resetExtractedTextFields(int id,Map values) {
		
		extractedTextFields.put(new Integer(id),values);
	}

	public void flushExtractedFields() {
		extractedTextFields.clear();
		extractedXMLFields.clear();
		
	}

    /**used to pass in actual data*/
    public void setMarkedContent(StringBuffer markedContent) {
        this.markedContent=markedContent;
        isMarked=true;
    }
}
